#!/usr/bin/env python
# life.py -- A curses-based version of Conway's Game of Life.
# Contributed by AMK
#
# An empty board will be displayed, and the following commands are available:
#  E : Erase the board
#  R : Fill the board randomly
#  S : Step for a single generation
#  C : Update continuously until a key is struck
#  Q : Quit
#  Cursor keys :  Move the cursor around the board
#  Space or Enter : Toggle the contents of the cursor's position
#
# TODO :
#   Support the mouse
#   Use colour if available
#   Make board updates faster
#

import random, string, traceback, re
import wrap
import generator
import cursesemu as curses
import tutorial
import asciianim

import pygame # Needed to get ticks

class UI:
  def __init__(self, stdscr):
    self.scr = stdscr
    stdscr.clear()
    
    self.uis = []
    self.pushUI( MainMenuUI(self, stdscr) )
    
    self.loop(stdscr)
    
  def loop(self, stdscr):
  
    while(1):
      c = stdscr.getch()                # Get a keystroke

      if 0<c:
        print c
        # Send the keystroke to the top UI
        self.uis[-1].sendKeyHit(c)

      self.uis[-1].tick( pygame.time.get_ticks() )

      stdscr.refresh()
      pass

  # Pops the current screen off of the stack, to exit the current screen and return control to the calling screen.
  # If autoQuit is true, then when the stack is empty, we will quit (or prompt the user to confirm a quit)
  def popUI( self, autoQuit = True ): 
    # Make sure that the display stack isn't empty...
    if ( len(self.uis) <= 1 ):
      # TODO: Push a "confirm quit?" dialog box onto the stack here.
      exit()
    else:
      self.uis[-1].hide()
      self.uis.pop()
      self.uis[-1].show()

  # Pushes a new screen onto the stack to relinquish control to this sub-menu
  def pushUI( self, newUI ):
    if len(self.uis) > 0:
      self.uis[-1].hide()
      
    self.uis.append( newUI )
    self.uis[-1].show()
  
  # Replaces the current UI with a different one
  def replaceUI( self, newUI ):
    if ( len(self.uis) >= 1 ):
      topUI = self.uis.pop()
      topUI.hide()
    self.pushUI( newUI )



class UISubScreen:
  def __init__( self, ui, stdscr ):
    self.titleText = ""
    self.helpText = "Sorry, no help is available for this screen."
    self.ui = ui
    self.scr = stdscr
    
    # Special GUIs are things like dialog boxes, help screens, and option windows
    self.is_special = False
    
    self.setup()

  def setup(self):
    pass
    
  def show(self):
    pass
    
  def hide(self):
    pass
    
  def sendKeyHit(self, key):
    # These are universal key handlers that apply no matter what UI you're in. Handlers for keystrokes for individual screens are handled in the onKeyHit handler.
    
    oldNumUIs = len(self.ui.uis)
    
    if not self.is_special:
      if key == curses.KEY_F1:   # F1
        # Game Help
        if (type(self.ui.uis[-1]) is not HelpUI):
          self.ui.pushUI( HelpUI(self.ui, self.scr) )
      elif key == curses.KEY_F2: # F2
        # Game Options
        if (type(self.ui.uis[-1]) is not OptionsUI):
          self.ui.pushUI( OptionsUI(self.ui, self.scr) )
      elif key == curses.KEY_F3: # F3
        # Game Quit
        if (type(self.ui.uis[-1]) is not QuitUI):
          self.ui.pushUI( QuitUI(self.ui, self.scr) )
          
    if key == 27: # ESC
      self.ui.popUI()
    
    # HACK
    # If we haven't pushed or popped any UIs...
    if (len(self.ui.uis) == oldNumUIs):
      # Then assume it's safe to pass the keystroke down to the main UI
      self.onKeyHit(key)
    
  def onKeyHit(self, key):
    pass
    
  def tick(self, ticks):
    pass
    
    
class HelpUI(UISubScreen):
  def setup(self):
    self.titleText = "Help"
    self.helpText = "Perhaps we put a fortune cookie in here. \n\nMaybe it help you. Maybe it don't."

    self.is_special = True
    self.win = self.scr.subwin(20, 60, 2, 10)
#    self.win = self.scr.subwin(30, 30, 5, 5)
 
  def show(self):
    self.scr.clear()

    self.win.box()
    self.scr.box()
    
    # Look at the help text of the window just above ours.
    curHelp = self.ui.uis[-2].helpText

    self.win.addstr(1,1, "Help: " + self.ui.uis[-2].titleText )
    
    self.win.addstr(18,2, "Press any key to continue...")
    
#    wrappedHelp = textwrap.wrap(curHelp, self.win.width()-2)
    wrappedHelp = wrap.wrap(curHelp, self.win.width()-4).split('\n')
     
    y = 3
    
    for line in wrappedHelp:
      self.win.addstr(y,2, line)
      y += 1
    
 
  def hide(self):
    self.scr.clear()   
    
  def onKeyHit(self, key):
    self.ui.popUI()
    

class MessageBoxUI(UISubScreen):
  def setup(self):
    self.titleText = "Message Box"
    self.messageText = "This space for rent."
    self.helpText = "This is a message box. Just hit a key or something, you'll be fine."

    self.is_special = True
    self.win = self.scr.subwin(10, 60, 5, 10)
 
  def show(self):
    self.scr.clear()

    self.win.box()
    self.scr.box()
    
    self.win.addstr(self.win.height()-2, 28, "[OK]")
    
#    wrappedHelp = textwrap.wrap(curHelp, self.win.width()-2)
    wrappedText = wrap.wrap(self.messageText, self.win.width()-4).split('\n')
     
    y = 2
    for line in wrappedText:
      self.win.addstr(y,2, line)
      y += 1
    
 
  def hide(self):
    self.scr.clear()   
    
  def onKeyHit(self, key):
    self.ui.popUI()

    
class MainMenuUI(UISubScreen):
  def setup(self):
    self.difficulty = 1
    self.lastTick = 0
    self.framedelay = 100
    self.trainY = self.scr.height() - 9
    self.trainX = self.scr.width() - 2
    self.trainFrame = 0
    
    self.win = self.scr.subwin(8, 20, 3, 5)
    self.titleText = "Main Menu"
    self.helpText = "All aboard! \n\nWelcome to the game! Here's the main menu -- think of it as a switching yard, where you can hop onto any rail in the system. If this is your first time playing the game, I recommend you punch your ticket for the [T]utorial by pressing the T key. Once you've gone through that, feel free to adjust your [D]ifficulty with the D key, or start a [N]ew game with the N key. Press Q to [Q]uit the game when you're done. You can press ESC to back out of a screen at any time (including this one).\n\nF1 always brings up help, F2 always brings up the options menu, and F3 always quits."
    
  def show(self):
    self.scr.clear()

    self.scr.box()
    self.win.box()
    self.refresh()
    
  def hide(self):
    self.win.clear()
    self.scr.clear()
  
  def onKeyHit(self, key):
    if (0<key<256):  
      key = chr(key)
    
      if key in 't':
        self.ui.pushUI( TutorialMenuUI(self.ui, self.scr) )
      if key in 'n' + chr(13):
        newGameUI = GameUI(self.ui, self.scr)
        newGameUI.problem = generator.generate_problem(self.difficulty)
        self.ui.pushUI( newGameUI )
      elif key in 'd':
        self.difficulty += 1
        if self.difficulty > 10:
          self.difficulty = 1
        self.refresh()
      elif key in 'q':
        self.ui.popUI()
      elif key in '1234567890':
        self.difficulty = int(key)
        if self.difficulty == 0:
          self.difficulty = 10
        self.refresh()

  def refresh(self):
    y = 3
    x = 21
    for line in asciianim.rexLogo:
      self.scr.addstr(y, x, line, curses.COLOR_BLUE + curses.A_BOLD)
      y += 1
  
  
    self.win.addstr(1,1,"[T]utorial", curses.COLOR_GREEN)
    self.win.addstr(2,1,"[D]ifficulty: [" + str(self.difficulty) + "] ", curses.COLOR_GREEN)
    self.win.addstr(3,1,"[N]ew game", curses.COLOR_GREEN)
    self.win.addstr(4,1,"[Q]uit", curses.COLOR_GREEN)
    self.win.addstr(6,1,"[F1] Help", curses.COLOR_GREEN + curses.A_BOLD)
    
    # Draw the train
    y = 0
    for line in asciianim.trainAnim[self.trainFrame]:
      wid = self.scr.width() - 2
      buff = " " * wid
      line = buff + line + buff
      
      line = line[self.trainX:self.trainX + wid]
      self.scr.addstr(self.trainY + y, 1, line, curses.COLOR_WHITE + curses.A_BOLD)
      y += 1
    
    
  def tick(self, ticks):
  
    wid = self.scr.width() - 2
  
    last_trainX = self.trainX
    last_trainFrame = self.trainFrame
      
    self.trainX = ((ticks / 150) % (wid * 2))
    self.trainFrame = ticks / 150 % len(asciianim.trainAnim)
    
    if (last_trainX != self.trainX) or (last_trainFrame != self.trainFrame):
      self.refresh()

class TutorialMenuUI(UISubScreen):
  def setup(self):
    self.titleText = "Tutorial Menu"
    self.helpText = "Welcome to Tutorial Train Tours! Please use the arrow keys or the letter designators to choose your tutorial. Once you've made your choice, press enter to confirm your ticket, and enjoy your trip on the RegEx Express!"
    
    self.selectionInx = 0
    self.tutorials = tutorial.load_tutorial()
    
    # Dimensions of the tutorial window are here. Height, width, top, left.
    self.win = self.scr.subwin(18, 70, 3, 5)
    
  def show(self):
    self.scr.clear()

    self.scr.box()
    self.win.box()
    
    self.refresh()

    
  def hide(self):
    self.win.clear()
    self.scr.clear()
  

  def onKeyHit(self, key):
    if key == 13:
      tut = GameUI(self.ui, self.scr)
      tut.problem = self.tutorials[self.selectionInx]
      self.ui.pushUI( tut )

      # Increment the selection index, so that when we come back, we are looking at the next tutorial, assuming they passed the last one.
      # Actually, in practice, it doesn't work very well, so commenting it out now.
      # self.selectionInx = (self.selectionInx + 1) % len(self.tutorials)

    else:
      if key == curses.KEY_UP:
        self.selectionInx -= 1
        if self.selectionInx < 0:
          self.selectionInx = len(self.tutorials) - 1
      elif key == curses.KEY_DOWN:
        self.selectionInx = (self.selectionInx + 1) % len(self.tutorials)
      elif 0<key<255:
        if chr(key) in 'abcdefghijklmnopqrstuvwxyz':
          newInx = key - ord('a')
          if newInx < len(self.tutorials):
            self.selectionInx = newInx
          
      self.refresh()

  def refresh(self):
    self.win.addstr(1, 1, "Please select tutorial with arrow keys or letters and press enter.")  
    
    cnt = 0
    # Draw all of the menu items
    for prob in self.tutorials:
      self.win.addstr(3+(cnt % (self.win.height()-6)), 2 + (int(cnt / (self.win.height() -6) ) * self.win.width() / 3), " [" + chr(cnt + ord("A")) + "] " + prob.title, curses.COLOR_GREEN)
      cnt += 1
   
   
    
    # Draw the currently-hilighted menu item
    cnt = self.selectionInx
    self.win.addstr(3+(cnt % (self.win.height()-6)), 2 + (int(cnt / (self.win.height() -6) ) * self.win.width() / 3), "*[" + chr(cnt + ord("A")) + "] " + self.tutorials[cnt].title, curses.COLOR_GREEN + curses.A_BOLD)

    self.win.addstr(self.win.height()-2, 2, "Press [F1] for help at any time in the game.", curses.COLOR_WHITE + curses.A_BOLD)

#    self.win.addstr(3+self.selectionInx, 2, "*" + str(self.selectionInx) + ") " + self.tutorials[self.selectionInx].title, curses.COLOR_GREEN + curses.A_BOLD)





class GameUI(UISubScreen):
  def setup(self):
    self.titleText = "Game"
    self.helpText = "The goal of the game is to create a regular expression string that will match on all of the words in the left column (should should all hilight in green), while failing on all of the strings in the right column (meaning they will all be hilighted in red). Type your regex with the keyboard (it will show up in the lower textbox), and press 'enter' when you want to submit your answer. You may press 'ESC' to cancel the current game."
    self.currentRegex = ""
    
    self.problem = None
  
    self.textBox  = self.scr.subwin(3, 80, 21, 0)
    self.instrBox = self.scr.subwin(22, 40, 0, 0)
    self.lWordBox = self.scr.subwin(22, 21, 0, 39)
    self.rWordBox = self.scr.subwin(22, 21, 0, 59)
        
  def show(self):
    self.scr.clear()
    
    self.scr.box()
    self.textBox.box()
    self.instrBox.box()
    self.lWordBox.box()
    self.rWordBox.box()
    
    self.lWordBox.addstr(1,1,"Words TO pass:")
    self.rWordBox.addstr(1,1,"Words to NOT:")
    self.instrBox.addstr(1,1,"Instructions go here.")
    
    self.textBox.addstr(1,1, "Your regex entry goes here")
    
    self.refresh()
    
  def hide(self):
    self.scr.clear()
  
  def onKeyHit(self, key):
    if key == 127 or key == 8: # Backspace or delete
      if len(self.currentRegex) > 0:
        # Trim a character off of the end
        self.currentRegex = self.currentRegex[0:-1]

    elif 32<=key<=126: # If the key is a valid ascii code
      self.currentRegex += chr(key)
    elif key == 13: # Carriage return
#      self.commitRegex()
      res = self.problem.get_result(self.currentRegex)

      msgBox = MessageBoxUI(self.ui, self.scr)

      if res.success == True:
        msgBox.messageText = "Congratulations, you passed the problem!"        
        self.ui.replaceUI(msgBox)
      else:
        msgBox.messageText = "Sorry, that regex doesn't solve the problem.\nYou must match (hilight in green) all of the words in the left column, while avoiding matching (keeping them all red) the words in the right column. You may press F1 for more help."
        self.ui.pushUI(msgBox)
        self.currentRegex = ""
  
    if self.ui.uis[-1] == self:
      self.refresh()
    
  def displayHilightedString( self, win, y, x, text, rgx ):
    # If the regex matches on the text
    if (rgx.search(text)):
      # First, add the whole string in a somewhat bland green
      win.addstr( y, x, text, curses.COLOR_GREEN )

      iter = rgx.finditer(text)
      for match in iter:
        # Then iterate through all of the matches, re-adding the text as bolded green to help it stand out for the user.
        win.addstr( y, x + match.span()[0], text[match.span()[0]:match.span()[1]], curses.COLOR_GREEN + curses.A_BOLD )
#      print text, "passes!"
    else:
#      print text, "fails!"
      win.addstr( y, x, text, curses.COLOR_RED + curses.A_BOLD )
  
  
    
  def refresh(self):
    self.textBox.clear()
    self.textBox.box()

    self.instrBox.clear()
    cnt = 0
    wrappedText = wrap.wrap(self.problem.explanation, self.instrBox.width()-4).split('\n')
     
    y = 2
    for line in wrappedText:
      self.instrBox.addstr(y,2, line)
      y += 1

    
    self.instrBox.addstr

    self.lWordBox.addstr(1,1,"Words TO pass:")
    self.rWordBox.addstr(1,1,"Words to NOT:")

    rgx = None
    try:
      rgx = re.compile(self.currentRegex)
#      print "Valid regex!"
      
      self.textBox.addstr( 1, 1, self.currentRegex, curses.COLOR_BLUE + curses.A_BOLD )
    
      y = 3
      for word in self.problem.goodwords:
        self.displayHilightedString( self.lWordBox, y, 1, word, rgx )
        y += 1
        
      y = 3
      for word in self.problem.badwords:
        self.displayHilightedString( self.rWordBox, y, 1, word, rgx )
        y += 1
    
    except:
#      print "Not a valid regex."
      self.textBox.addstr( 1, 1, self.currentRegex, curses.COLOR_RED + curses.A_BOLD )
      
      y = 3
      for word in self.problem.goodwords:
        self.lWordBox.addstr( y, 1, word )
        y += 1
        
      y = 3
      for word in self.problem.badwords:
        self.rWordBox.addstr( y, 1, word )
        y += 1

    self.scr.refresh()


    

def main(stdscr):
    ui = UI(stdscr)                    # Enter the main loop

def run():
    curses.wrapper(main)

if __name__ == '__main__':
    curses.wrapper(main)
